from collections import namedtuple

from nagiosplugin import (
    Resource,
    Check,
    Metric,
    ScalarContext,
    guarded,
    Context,
    Summary,
    Ok, Warn, Critical, Unknown
)


class StateContext(Context):
    """
    Nagios context class for data that contains a state.
    """

    def __init__(self, name, **kwargs):
        fmt_metric = u'{name}: {value[1]}'
        super(StateContext, self).__init__(
            name,
            fmt_metric=fmt_metric,
            **kwargs
        )

    def evaluate(self, metric, resource):
        """Evaluate the result based on its state."""
        (check_state, text) = metric.value
        if check_state == 'ok':
            state = Ok
        elif check_state == 'warning':
            state = Warn
        elif check_state == 'critical':
            state = Critical
        elif check_state == 'unknown':
            state = Unknown
        else:
            raise ValueError(u'{} is not a supported value'.format(
                check_state
            ))
        return self.result_cls(state, '', metric)


class DetailedContext(ScalarContext):
    """
    Nagios context class for data that has additional information in the
    `detail` attribute.
    """
    def describe(self, metric):
        return u'{name}: {value}{uom} ({detail})'.format(
            name=metric.name,
            value=metric.value,
            uom=metric.uom,
            detail=metric.detail
        )


class DetailedMetric(
    namedtuple('DetailedMetric', Metric._fields + ('detail',)),
    Metric
):
    """
    Nagios metric class for data that has additional information in the
    `detail` attribute.
    """
    def __new__(
        cls, name, value, uom=None, min=None, max=None, context=None,
        contextobj=None, resource=None, detail=None
    ):
        return tuple.__new__(
            cls,
            (
                name, value, uom, min, max, context or name, contextobj,
                resource, detail
            )
        )


class CustomSummary(Summary):
    """
    Nagios summary class that implements custom output formatting.
    """

    def __init__(self, verbose):
        self.verbosity_level = verbose
        super(CustomSummary, self).__init__()

    def ok(self, results):
        """Format an OK result."""
        result_count = len(results.results)
        if result_count == 1:
            return u'{}'.format(results[0])
        else:
            return u'all {} results are ok'.format(result_count)

    def problem(self, results):
        """Format a non-OK result."""
        result_count = len(results.results)
        if result_count == 1:
            return str(results.first_significant)
        return u'{}: {}'.format(
            u', '.join(
                u'{}/{} {}'.format(len(v), result_count, k)
                for k, v in sorted(results.by_state.items())
            ),
            str(results.first_significant)
        )

    def verbose(self, results):
        """Print additional output based on our verbosity level."""
        msgs = []
        if not results:
            return msgs
        # If all checks are Ok, require verbose level 2 to print them all out
        if self.verbosity_level == 1 and results.most_significant_state == Ok:
            pass
        elif self.verbosity_level <= 2:
            for result in results.by_state[results.most_significant_state]:
                msgs.append(u'{0}: {1}'.format(result.state, result))
        elif self.verbosity_level == 3:
            for severity in [Critical, Warn, Unknown, Ok]:
                for result in results.by_state.get(severity, []):
                    msgs.append(u'{0}: {1}'.format(result.state, result))
        return msgs


# This camel snake case is on purpose. Naming this class seems to
# be the only way to make the output as wanted.
class Check_Vmware(Resource):
    """
    Nagios resource class for performance data.
    """

    def __init__(self, resources, query_params):
        self.resources = resources
        self.query_params = query_params
        return

    def probe(self):
        """Query the resource for performance data."""
        data = self.resources.query(self.query_params)
        for result in data['data']:
            if 'state' in result:
                yield Metric(
                    result['name'],
                    (result['state'], result['value']),
                    context='state'
                )
            elif 'detail' in result:
                yield DetailedMetric(
                    result['name'],
                    result['value'],
                    uom=result.get('unit', None),
                    context='detail',
                    detail=result['detail']
                )
            else:
                yield Metric(
                    result['name'],
                    result['value'],
                    uom=result.get('unit', None),
                    context='scalar'
                )


@guarded(verbose=0)
def check(resources, query_params, warning, critical, verbose):
    """
    Setup a Nagios check for the given parameters.
    """
    check = Check(
        Check_Vmware(resources, query_params),
        ScalarContext(
            'scalar',
            warning=warning,
            critical=critical
        ),
        DetailedContext(
            'detail',
            warning=warning,
            critical=critical
        ),
        StateContext('state'),
        CustomSummary(verbose=verbose)
    )
    check.main(verbose=verbose)
