import ssl
from functools import wraps

from atexit import register as atexit
from threading import Lock

from pyVim.connect import Disconnect, SmartConnect
from pyVmomi import vim, vmodl, SoapStubAdapter

from check_vmware.vsphere.constants import (
    API_TYPE_HOST,
    API_TYPE_VCENTER
)
from check_vmware.vsphere.exceptions import (
    VsphereException,
    VsphereConnectionError,
    VsphereInvalidLogin,
    translate_exceptions
)
from check_vmware.vsphere import query
from check_vmware.vsphere.list_counters import list_counters, list_all_counters
try:
    from http.client import RemoteDisconnected
except ImportError:
    try:
        from http.client import BadStatusLine as RemoteDisconnected
    except ImportError:
        from httplib import BadStatusLine as RemoteDisconnected


def retry(f):
    """
    Decorator that retries the function call in case of a connection error.
    """
    @wraps(f)
    def wrapper(connection, *args, **kwargs):
        try:
            return f(connection, *args, **kwargs)
        except (RemoteDisconnected, vim.fault.NotAuthenticated):
            connection.reconnect()
            return f(connection, *args, **kwargs)
    return wrapper


class VsphereConnection(object):
    """
    Connection to the vSphere API. Stores the ServiceInstance and
    ServiceInstanceContent objects for quicker access.
    """

    def __init__(self, host, port, user, password, ignore_ssl):
        """
        Connect to a vSphere server and set up a ServiceInstance object.
        """
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.ignore_ssl = ignore_ssl
        self.cache_key = host
        self.host_type = None
        self._vsan_stub = None
        self._connect()

    def _connect(self):
        try:
            self._vsan_stub = None
            ssl_context = self._ssl_context()
            si = SmartConnect(
                host=self.host,
                port=self.port,
                user=self.user,
                pwd=self.password,
                connectionPoolTimeout=-1,
                sslContext=ssl_context
            )
            atexit(Disconnect, si)
            self.si = si
            self.content = si.RetrieveContent()
            self.host_type = self.content.about.apiType
        except vim.fault.InvalidLogin as e:
            raise VsphereInvalidLogin(self.user)
        except vmodl.fault.HostCommunication as e:
            raise VsphereConnectionError(e.msg)
        except IOError as e:
            raise VsphereConnectionError(str(e))

    def reconnect(self):
        """Replace the underlying ServiceInstance object with a new one."""
        self._connect()

    @property
    def stub(self):
        """
        Property attribute for the SessionOrientedStub of our ServiceInstance.
        """
        return self.si._GetStub()

    @property
    def vsan_stub(self):
        """
        Property attribute for the vsan SessionOrientedStub of our
        ServiceInstance.
        """
        if self._vsan_stub:
            return self._vsan_stub
        if self.host_type == API_TYPE_VCENTER:
            path = '/vsanHealth'
        elif self.host_type == API_TYPE_HOST:
            path = '/vsan'
        else:
            raise VsphereException(
                u'Unrecognized API type: {}'.format(self.host_type)
            )
        ssl_context = self._ssl_context()
        # TODO: Automatically detect this
        version = 'vim.version.version10'
        self._vsan_stub = SoapStubAdapter(
            host=self.host,
            path=path,
            version=version,
            sslContext=ssl_context
        )
        self._vsan_stub.cookie = self.stub.cookie
        return self._vsan_stub

    def _ssl_context(self):
        return ssl._create_unverified_context() if self.ignore_ssl else None

    @retry
    @translate_exceptions
    def query(
            self,
            counter_name,
            entity_path,
            interval,
            instance,
            flags
    ):
        """Query the vSphere API for performance data."""
        return query.query(
            self,
            counter_name,
            entity_path,
            interval,
            instance,
            flags
        )

    @retry
    @translate_exceptions
    def list_counters(self, entity_path, interval):
        """Query the vSphere API for all available counters."""
        return list_counters(
            self,
            entity_path=entity_path,
            interval=interval
        )

    @retry
    @translate_exceptions
    def list_all_counters(self):
        """Query the vSphere API for all known counters."""
        return list_all_counters(self)


class ConnectionManager(object):
    """Manager that recycles existing connections to the same host and user."""

    def __init__(self):
        self._connections = {}
        self._lock = Lock()

    def get_connection(
        self,
        host,
        port,
        user,
        password,
        ignore_ssl,
        force_connect=False
    ):
        """
        Return an active VsphereConnection for the given vSphere server
        and user. Re-use previously created instance if present.
        """

        with self._lock:
            key = u'{}:{}@{}:{}:{}'.format(
                user,
                password,
                host,
                port,
                ignore_ssl
            )
            if key in self._connections and force_connect is False:
                return self._connections[key]
            connection = VsphereConnection(
                host,
                port,
                user,
                password,
                ignore_ssl
            )
            self._connections[key] = connection
            return connection


_connection_manager = ConnectionManager()


def connect(host, port, user, password, ignore_ssl=False, force_connect=False):
    """
    Return an active VsphereConnection for the given vSphere server
    and user. Re-use previously created instance if present.
    """

    return _connection_manager.get_connection(
        host,
        port,
        user,
        password,
        ignore_ssl,
        force_connect=force_connect
    )
